Option Explicit ' rg_itree.vbs, Rick G.
' Purpose  : Build NWC2 Instrument Tree
' Syntax   : WScript rg_ITree.vbs <file> (*.dls, *.sf2, *.sbk)
' References:
' RIFF File format
'   http://msdn.microsoft.com/en-us/library/ms713231(VS.85).aspx
' General MIDI Level 1 Sound Set
'   www.midi.org/techspecs/gm1sound.php
' Downloadable Sounds Level 1
'   www.midi.org/techspecs/dls/dsl1v11b.pdf
' Downloadable Sounds Level 2
'   www.amei.or.jp/report/AMEI_MMA/dls2v10a.pdf
' SoundFont Technical Specification
'   http://connect.creativelabs.com/developer/SoundFont/sfspec21.pdf
Const version = "1.00"
Dim fso: Set fso = CreateObject("Scripting.FileSystemObject")
Dim wso: Set wso = CreateObject("WScript.Shell")
Dim file: For Each file In WScript.Arguments: Exit For: Next
'file = "C:\Program Files\SynthFont\SYNTHGMS.SF2"
If Not fso.FileExists(file) Then file = SyntaxTest
If Not fso.FileExists(file) Then WScript.Quit(99)

Dim aClan: aClan = Split("\GM1\|\GM2\|\Variations\|\Drum Kits", "|")
'aClan = Split(" | | |\Drum Kits", "|")
'aClan = Split("\|\|\|\Drum Kits", "|")
Dim aFam: aFam = Split("Piano|Chromatic Percussion|Organ" _
	& "|Guitar|Bass|Strings|Ensemble|Brass|Reed|Pipe|Synth Lead" _
	& "|Synth Pad|Synth Effects|Ethnic|Percussive|Sound Effects", "|")
Dim tag, typRIFF, typLIST, size, data ' RIFF vars
Dim ulBank, ulInstrument ' DLS vars
Dim sName, patch, msb, lsb ' NWC2 itree vars
Dim isDrum, sList ' sortable patch list
Dim tso: Set tso = fso.OpenTextFile(file)
Do ' until Exit Do
	tag = GetChunkType(tso)
	If Len(tag) Then size = cvii(tso.Read(4)) Else Exit Do
	Select Case tag
		Case "RIFF": typRIFF = tso.Read(4)
		Case "LIST": typLIST = tso.Read(4)
			If typLIST = "ins " Then AddLine ' new DLS instrument
		Case "insh" ' DLS: Instrument Header
			data = tso.Read(size)
			ulBank = cvii(Mid(data, 5, 4))
			ulInstrument = cvii(Mid(data, 9, 4))
			isDrum = CBool(ulBank And &h80000000) ' bit31 = 1
			patch = ulInstrument And 127 ' bits 0-6
			msb = ulBank \ 256 And 127 ' bits 8-14
			lsb = ulBank And 127 ' bits 0-6
		Case "INAM" ' INFO-List Sub-chunk
			data = tso.Read(size)
			sName = Left(data, size -1)
		Case "ptbl" ' DLS: Pool Table
			AddLine: Exit Do ' last DLS instrument
		Case "phdr" ' sfspec21.pdf: 7.2 The PHDR Sub-chunk
			data = tso.Read(size)
			Do While Len(data) > 38 ' last record is dummy
				sName = Left(data, 20) ' achPresetName[20]
				patch = CLng(cvii(Mid(data, 21, 2))) ' wPreset
				msb = CLng(cvii(Mid(data, 23, 2))) ' wBank
				lsb = 0
				sName = Split(sName & Chr(0), Chr(0))(0) ' deZ
				isDrum = msb = 128: If isDrum Then msb = 0
				AddLine
				data = Mid(data, 39) ' next sfPresetHeader
			Loop
			Exit Do
		Case Else: tso.Skip(size) ' to end of chunk
	End Select
	tso.Skip(size And 1) ' RIFF files are word aligned
Loop
tso.Close
Dispatch file, MakeTree(ShellSort(Split(Mid(sList, 2), vbLf), 1))

Sub AddLine ' append sortable line to sList
	Dim a, n: a = Array(0, patch, msb, lsb, sName)
	If Len(patch) Then ' must have patch
		If isDrum Then
			a(0) = UBound(aClan)
		ElseIf lsb = 120 Then ' msb??
			a(0) = 1 ' GM2
		ElseIf msb Or lsb Then
			a(0) =  2 ' Variations (GS, XG, ...)
		End If ' else zero (GM1)
		For n = 1 To 3: a(n) = Right("0000" & a(n), 5): Next
		sList = sList & vbLf & Join(a, vbTab) ' Tab delimited
	End If
	patch = "": msb = "": lsb = "": sName = "" ' invalidate
End Sub

Function MakeTree(aTSV) ' array of Tab Separated Values
	Const test = "=0123456789abcdefghijklmnopqrstuvwxyz"
	Dim d: Set d = CreateObject("Scripting.Dictionary")
	Dim a, e, s, sBranch, isGM1: isGM1 = True ' assume
	s = "\=Bank(0,0)" ' Instrument Tree default
	For Each e In aTSV
		a = Split(e, vbTab)
		' Make Instument Name nwcitree compatible
		a(4) = RTrim(Replace(a(4), "=", "-")) ' must remove
		Select Case InStr(1, test, Left(a(4), 1), vbTextCompare)
			Case 0: a(4) = "o" & a(4) ' must begin with letter or digit
			Case 1: a(4) = "Unnamed" ' cannot be empty
		End Select
		If d.Exists(a(4)) Then ' name exists
			d(a(4)) = d(a(4)) + 1
			a(4) = a(4) & "#" & d(a(4))
		Else
			d.Add a(4), 0
		End If
		' add branch if it has changed
		a(0) = vbCrLf & vbCrLf & aClan(a(0))
		If Right(a(0), 1) = "\" Then a(0) = a(0) & aFam(a(1) \ 8)
		If sBranch <> a(0) Then s = s & a(0): sBranch = a(0)
		' add patch
		s = s & vbCrLf & a(4) & "=" & CLng(a(1)) ' Name=Patch
		If a(2) Or a(3) Then ' add Bank(msb,lsb)
			s = s & ",Bank(" & CLng(a(2)) & "," & CLng(a(3)) & ")"
			isGM1 = False ' not General Midi 1
		End If
	Next
	a = Array(vbCrLf & aClan(0), vbCrLf & "\")
	If isGM1 Then s = "# " & Replace(s, a(0), a(1)) ' simplify
	MakeTree = s
End Function ' nwcitree out

Function SyntaxTest
	Dim e, user, buttons: buttons = vbInformation
	Dim title: title = WScript.ScriptName & " Version " & version
	Dim prompt: prompt = "Syntax is:" & vbLf & "  " _
		& WScript.ScriptName & " <file>  (*.dls | *.sf2)" & vbLf
	For Each e In fso.GetFolder(fso.BuildPath( _
		fso.GetSpecialFolder(0), "system32\drivers")).Files
		If LCase(fso.GetExtensionName(e)) = "dls" Then Exit For
	Next
	If IsObject(e) Then
		buttons = vbOKCancel Or vbQuestion
		prompt = prompt & vbLf & "Test with: " & e.Name & " ?"
	End If
	user = MsgBox(prompt, buttons, title)
	If user = vbOK And IsObject(e) Then SyntaxTest = e.Path
End Function ' test file out or Empty

Sub Dispatch(srcFile, contents)
	Dim buttons, user, prompt, a(2), n
	Dim f: f = fso.GetExtensionName(srcFile) _
		& "_" & fso.GetBaseName(srcFile) & ".nwcitree"
	Dim count: count = UBound(Split(contents, "=")) -1
	Select Case count
		Case 0
			buttons = vbExclamation Or vbSystemModal
			prompt = "No instruments found in: " & srcFile
			MsgBox prompt, buttons, WScript.ScriptName
			Exit Sub ' nothing to save or display
		Case 1
			prompt = "1 instrument found in: " & srcFile
		Case Else
			prompt = count & " instruments found in: " & srcFile
	End Select
	buttons = vbYesNoCancel Or vbQuestion Or vbSystemModal
	' array of folders, 0=user, 1=NWC2 default, 2=%TEMP%
	a(0) = 	RegData("HKCU\Software\NoteWorthy Software\" _
		& "NoteWorthy Composer 2\Folders\InstrumentTrees")
	a(1) = 	fso.BuildPath(RegData("HKLM\Software\" _
		& "Microsoft\Windows\CurrentVersion\Uninstall\" _
		& "NoteWorthy Composer 2\InstallLocation"), "itrees")
	a(2) = fso.GetSpecialFolder(2)
	n = 0: Do Until fso.FolderExists(a(n)): n = n + 1: Loop
	If n < 2 Then ' NWC2 found
		If fso.FileExists(fso.BuildPath(a(n), f)) Then
			prompt = prompt & vbLf & vbTab _
				& "This Instrument Tree exists. Overwrite?"
			buttons = buttons Or vbDefaultButton2
		Else
			prompt = prompt & vbLf & vbTab _
				& "Save this Instrument Tree?"
		End If
		user = MsgBox(prompt, buttons, f)
	End If
	If user = vbCancel Then Exit Sub ' discard
	If user <> vbYes Then n = 2
	If n = 2 Then f = f & ".txt"
	f = fso.BuildPath(a(n), f)
	fso.CreateTextFile(f, True).Write contents
	If n < 2 Then Exit Sub ' user save, don't display
	wso.Run(Chr(34) & f & Chr(34)) ' display
	WScript.Sleep 5000: fso.DeleteFile(f) ' cleanup
End Sub ' contents saved, displayed or discarded

Function RegData(sKey)
	On Error Resume Next: RegData = wso.RegRead(sKey)
End Function ' data out (or Empty if sKey not found)

Function GetChunkType(o) ' TestStream Object in
	On Error Resume Next: GetChunkType = o.Read(4)
End Function ' fourCC out (or Empty if err)

Function cvii(s) ' Intel bytes in (LittleEndian)
	Dim n
	For n = 1 To Len(s)
		cvii = Right("0" & Hex(Asc(Mid(s, n))), 2) & cvii
	Next
	cvii = "&h0" & cvii
End Function ' HexStr out

Function ShellSort(a, n) ' array, 1 (ascend) -1 (descend)
	Dim i, j, r, Swap: r = a: j = UBound(r)
	Do While j: j = j \ 2
		Do: Swap = vbFalse
			For i = 0 To UBound(r) - j
				If StrComp(r(i), r(i + j)) = n Then
					Swap = r(i): r(i) = r(i + j)
					r(i + j) = Swap: Swap = vbTrue
				End If
	Next:	Loop While Swap: Loop: ShellSort = r
End Function	' sorted array out, vbBinaryCompare